﻿//Programmer: Blaine (YouTube: physicsguy2011)

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace gcode_music
{
    public partial class Form1 : Form
    {
        

        //C0 to B0
        double [] frequencies = {16.351, 17.324, 18.354, 19.445, 20.601, 21.827, 23.124, 
                                   24.499, 25.956, 27.5, 29.135, 30.868};

        //convert the selected key to an index for the frequencies lookup table
        int identifyKey(string note)
        {
            switch (note){
                case "C":
                    return 0;
                
                case "C#/Db":
                    return 1;
                    
                case "D":
                    return 2;
                    
                case "D#/Eb":
                    return 3;
                    
                case "E":
                    return 4;

                case "F":
                    return 5;

                case "F#/Gb":
                    return 6;

                case "G":
                    return 7;

                case "G#/Ab":
                    return 8;

                case "A":
                    return 9;

                case "A#/Bb":
                    return 10;

                case "B":
                    return 11;
            }
            return 0;
        }

        public Form1()
        {
            InitializeComponent();
        }

        //Lock or unlock the settings options when the checked state is changed for the lock check box
        private void chkLock_CheckedChanged(object sender, EventArgs e)
        {
            if (chkLock.CheckState == CheckState.Checked)
            {
                txtSteps.Enabled = false;
                cbBaseNote.Enabled = false;
                nudBPM.Enabled = false;
                txtMaxZ.Enabled = false;
            }
            else
            {
                txtSteps.Enabled = true;
                cbBaseNote.Enabled = true;
                nudBPM.Enabled = true;
                txtMaxZ.Enabled = true;
            }
        }

        //add the note per the note settings controls to the list view
        private void btnAdd_Click(object sender, EventArgs e)
        {
            string [] subitems = {cbOctave.Text, cbNote.Text, chkDotted.Checked.ToString()};

            listViewNotes.Items.Add(cbKey.Text).SubItems.AddRange(subitems);
        }

        //move selected not in list view up one
        private void btnUp_Click(object sender, EventArgs e)
        {

            ListViewItem temp = listViewNotes.SelectedItems[0];
            int tempIndex = temp.Index;

            if (tempIndex > 0)
            {
                listViewNotes.Items.RemoveAt(tempIndex);
                listViewNotes.Items.Insert(tempIndex - 1, temp);
            }
        }

        //move selected note in list view down one
        private void btnDown_Click(object sender, EventArgs e)
        {
            ListViewItem temp = listViewNotes.SelectedItems[0];
            int tempIndex = temp.Index;

            if (tempIndex < listViewNotes.Items.Count - 1)
            {
                listViewNotes.Items.RemoveAt(tempIndex);
                listViewNotes.Items.Insert(tempIndex + 1, temp);
            }
        }

        //delete selected note from list view
        private void btnRemove_Click(object sender, EventArgs e)
        {
            if (listViewNotes.SelectedItems.Count > 0)
            {
                listViewNotes.Items.RemoveAt(listViewNotes.SelectedItems[0].Index);
            }
        }

        //set defaults on form load
        private void Form1_Load(object sender, EventArgs e)
        {
            cbKey.SelectedIndex = 0;
            cbOctave.SelectedIndex = 4;
            cbNote.SelectedIndex = 2;
            cbBaseNote.SelectedIndex = 2;
        }

        //generate the music GCode
        private void btnGenerate_Click(object sender, EventArgs e)
        {
            double steps = 0.0;
            double baseNote = Convert.ToDouble(cbBaseNote.Text);
            double BPM = Convert.ToDouble(nudBPM.Value);
            double tempFeedrate = 0.0;
            double zTravel = 10.0;
            double zTravelLast = 0.0;
            int zTravelDir = 0;
            int restTime = 0;
            double incZTravel = 0.0;
            double incZTravelDotted = 0.0;
            double note = 0.0;
            double octave = 0.0;
            string key = "";
            string dotted = "";
            string prevNote = "";
            string curNote = "";

            try
            {
                txtSteps.BackColor = Color.White;
                steps = Math.Abs(Convert.ToDouble(txtSteps.Text));
            }
            catch (System.FormatException fe)
            {
                txtSteps.BackColor = Color.Red;
                txtSteps.SelectAll();
                txtSteps.Focus();
                return;
            }

            txtMaxZ.BackColor = Color.White;

            rtxtGcode.Text = "";
            rtxtGcode.Text += "G28\n"; //home machine first
            rtxtGcode.Text += "G1 F5000 Z10\n"; //go to lowest Z which is 10mm above bed
            rtxtGcode.Text += "G4 S1\n"; //dramatic pause

            for (int i = 0; i < listViewNotes.Items.Count; i++)
            {
                note = Convert.ToDouble(listViewNotes.Items[i].SubItems[2].Text); //note type. quarter, half , whole, etc....
                octave = Convert.ToDouble(listViewNotes.Items[i].SubItems[1].Text); 
                key = listViewNotes.Items[i].SubItems[0].Text; //key. C, D, A#, etc....
                dotted = listViewNotes.Items[i].SubItems[3].Text; //add note length + half the note length if true

                //will crash if i = 0
                if (i > 0)
                {
                    //holds previous note for comparison to current note (curNote) later on
                    prevNote = listViewNotes.Items[i - 1].SubItems[0].Text + listViewNotes.Items[i - 1].SubItems[1].Text;
                }
                curNote = listViewNotes.Items[i].SubItems[0].Text + listViewNotes.Items[i].SubItems[1].Text;

                //adds a rest to GCode if key is set to "R"
                if (listViewNotes.Items[i].SubItems[0].Text == "R")
                {
                    restTime = (int)Math.Round((baseNote / note) * (1 / BPM) * 60000);
                    rtxtGcode.Text += "G4 P" + restTime + "\n";
                }
                else
                {
                    
                    //determine required feed rate for desired frequency
                    //math: (note frequency * 60 seconds per minute * 2^octave) / steps per mm
                    //units: (steps/s) * (s/min) / (steps/mm) = (steps/s) * (s/min) * (mm/steps) = (steps/min) * (mm/steps) = mm/min
                    tempFeedrate = (frequencies[identifyKey(key)] * 60 * Math.Pow(2, octave)) / steps;

                    //save previous z position for later use
                    zTravelLast = zTravel;

                    //incremental z moves. First for when dotted is not checked. second for when dotted is checked
                    incZTravel = (baseNote / note) * (1 / BPM) * tempFeedrate;
                    incZTravelDotted = (baseNote / note + baseNote / note / 2) * (1 / BPM) * tempFeedrate;

                    //if z travel direction is up
                    if (zTravelDir == 0)
                    {
                        if (dotted == "True")
                        {
                            zTravel += incZTravelDotted;
                        }
                        else
                        {
                            zTravel += incZTravel;
                        }
                        try
                        {
                            //reverse direction if max z height is reached
                            if (zTravel > Convert.ToDouble(txtMaxZ.Text))
                            {
                                zTravelDir = 1;
                                zTravel = zTravelLast;
                                if(dotted == "True"){
                                    zTravel -= incZTravelDotted;
                                }
                                else{
                                    zTravel -= incZTravel;
                                }
                            }
                        }
                        catch
                        {
                            txtMaxZ.BackColor = Color.Red;
                            txtMaxZ.SelectAll();
                            txtMaxZ.Focus();
                            return;
                        }
                    }
                    else
                    {
                        if (dotted == "True")
                        {
                            zTravel -= incZTravelDotted;
                        }
                        else
                        {
                            zTravel -= incZTravel;
                        }
                        //reverse direction if min z height is reached (10mm above bed)
                        if (zTravel < 10)
                        {
                            zTravelDir = 0;
                            zTravel = zTravelLast;
                            if (dotted == "True")
                            {
                                zTravel += incZTravelDotted;
                            }
                            else
                            {
                                zTravel += incZTravel;
                            }
                        }
                    }

                    //add a brief pause between consecutive notes of that are the same key and octave
                    //otherwise the notes will run together and sound like one long note
                    if (curNote == prevNote)
                    {
                        rtxtGcode.Text += "G4 P1\n";
                    }

                    //generate single note gcode
                    rtxtGcode.Text += "G1 F" + Math.Round(tempFeedrate, 3).ToString() + " Z" + Math.Round(zTravel, 3).ToString() + "\n";

                    
                    
                }
            }

            rtxtGcode.Text += "G4 S1\n"; //pause to let the awesomeness sink in
            rtxtGcode.Text += "G28"; //last thing, home machine
        }

        //add a rest to list view
        private void btnRest_Click(object sender, EventArgs e)
        {
            string[] subitems = { cbOctave.Text, cbNote.Text, chkDotted.Checked.ToString() };

            listViewNotes.Items.Add("R").SubItems.AddRange(subitems);
        }

        //save GCode that has been generated
        private void btnSaveGCode_Click(object sender, EventArgs e)
        {
            if (sfdSaveGCode.ShowDialog() == System.Windows.Forms.DialogResult.OK &&
            sfdSaveGCode.FileName.Length > 0)
            {
                rtxtGcode.SaveFile(sfdSaveGCode.FileName, RichTextBoxStreamType.PlainText);
            }
        }

        //save note data to file
        private void btnSave_Click(object sender, EventArgs e)
        {
            string temp = "";

            if (sfdSave.ShowDialog() == System.Windows.Forms.DialogResult.OK &&
           sfdSave.FileName.Length > 0)
            {
                temp += "V1" + Environment.NewLine;
                temp += txtSteps.Text + "," + cbBaseNote.Text + "," + nudBPM.Value.ToString()
                        + "," + txtMaxZ.Text + Environment.NewLine;
                for (int i = 0; i < listViewNotes.Items.Count; i++)
                {
                    temp += listViewNotes.Items[i].SubItems[0].Text + "," + listViewNotes.Items[i].SubItems[1].Text
                        + "," + listViewNotes.Items[i].SubItems[2].Text + "," + listViewNotes.Items[i].SubItems[3].Text + Environment.NewLine;
                }
                using (StreamWriter outfile = new StreamWriter(sfdSave.FileName, false, Encoding.ASCII))
                {
                    outfile.Write(temp);
                }

            }
        }

        //load note data from file
        private void btnOpen_Click(object sender, EventArgs e)
        {
            if (ofdOpen.ShowDialog() == System.Windows.Forms.DialogResult.OK &&
           ofdOpen.FileName.Length > 0)
            {
                listViewNotes.Items.Clear();
                chkLock.Checked = true;
                string [] temp = File.ReadAllLines(ofdOpen.FileName);
                string[] version = temp[0].Split(',');
                string[] setup = temp[1].Split(',');
                txtSteps.Text = setup[0];
                cbBaseNote.Text = setup[1];
                nudBPM.Value = Convert.ToDecimal(setup[2]);
                txtMaxZ.Text = setup[3];
                for (int i = 2; i < temp.GetLength(0); i++)
                {
                    string[] noteParams = temp[i].Split(',');
                    listViewNotes.Items.Add(noteParams[0]);
                    listViewNotes.Items[i - 2].SubItems.Add(noteParams[1]);
                    listViewNotes.Items[i - 2].SubItems.Add(noteParams[2]);
                    listViewNotes.Items[i - 2].SubItems.Add(noteParams[3]);
                }
            }
        }

        //delete all notes in list view
        private void btnClear_Click(object sender, EventArgs e)
        {
            DialogResult res = MessageBox.Show("Are you sure you want to clear all notes?", "Clear All Notes", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning);
            if (res == DialogResult.OK)
            {
                listViewNotes.Items.Clear();
            }
        }
    }
}
